@@ -20,10 +20,12 @@ import (
 	"bytes"
 	"context"
 	"crypto/ecdsa"
+	"crypto/ed25519"
 	"crypto/elliptic"
 	"crypto/rand"
 	_ "embed"
 	"encoding/base64"
+	"encoding/pem"
 	"fmt"
 	"io"
 	"log/slog"
@@ -43,18 +45,19 @@ import (
 	apko_build "chainguard.dev/apko/pkg/build"
 	apko_types "chainguard.dev/apko/pkg/build/types"
 	apko_cpio "chainguard.dev/apko/pkg/cpio"
-	"chainguard.dev/melange/internal/logwriter"
-	"chainguard.dev/melange/pkg/license"
 	"github.com/chainguard-dev/clog"
-	"github.com/charmbracelet/log"
 	v1 "github.com/google/go-containerregistry/pkg/v1"
 	"github.com/google/go-containerregistry/pkg/v1/tarball"
 	"github.com/kballard/go-shellquote"
+	"github.com/u-root/u-root/pkg/cpio"
 	"go.opentelemetry.io/otel"
 	"golang.org/x/crypto/ssh"
 	"golang.org/x/crypto/ssh/agent"
 	"golang.org/x/crypto/ssh/knownhosts"
 	"golang.org/x/term"
+
+	"chainguard.dev/melange/internal/logwriter"
+	"chainguard.dev/melange/pkg/license"
 )
 
 var _ Debugger = (*qemu)(nil)
@@ -110,7 +113,8 @@ func (bw *qemu) Run(ctx context.Context, cfg *Config, envOverride map[string]str
 }
 
 func (bw *qemu) Debug(ctx context.Context, cfg *Config, envOverride map[string]string, args ...string) error {
-	clog.FromContext(ctx).Debugf("running debug command: %v", args)
+	log := clog.FromContext(ctx)
+	log.Debugf("running debug command: %v", args)
 
 	// default to root user, unless a different user is specified
 	user := "root"
@@ -119,7 +123,7 @@ func (bw *qemu) Debug(ctx context.Context, cfg *Config, envOverride map[string]s
 	}
 
 	log.Debug("qemu: ssh - get user ssh key pair")
-	pubKey, err := getUserSSHKey()
+	pubKey, err := getUserSSHKey(ctx)
 	if err != nil {
 		log.Warn("qemu: could not get user ssh key pair, using ephemeral ones")
 	}
@@ -146,14 +150,14 @@ func (bw *qemu) Debug(ctx context.Context, cfg *Config, envOverride map[string]s
 	fd := int(os.Stdin.Fd())
 	oldState, err := term.MakeRaw(fd)
 	if err != nil {
-		return fmt.Errorf("failed to set terminal to raw mode: %v", err)
+		return fmt.Errorf("failed to set terminal to raw mode: %w", err)
 	}
 	//nolint:errcheck
 	defer term.Restore(fd, oldState)
 
 	width, height, err := term.GetSize(fd)
 	if err != nil {
-		return fmt.Errorf("failed to get terminal size: %v", err)
+		return fmt.Errorf("failed to get terminal size: %w", err)
 	}
 
 	winch := make(chan os.Signal, 1)
@@ -383,6 +387,7 @@ func (bw *qemu) TerminatePod(ctx context.Context, cfg *Config) error {
 	defer os.Remove(cfg.ImgRef)
 	defer os.Remove(cfg.Disk)
 	defer os.Remove(cfg.SSHHostKey)
+	defer secureDelete(ctx, cfg.InitramfsPath)
 
 	clog.FromContext(ctx).Info("qemu: sending shutdown signal")
 	err := sendSSHCommand(ctx,
@@ -451,7 +456,7 @@ func (bw *qemu) WorkspaceTar(ctx context.Context, cfg *Config, extraFiles []stri
 	// we append also all the necessary files that we might need, for example Licenses
 	// for license checks
 	if len(extraFiles) > 0 {
-		retrieveCommand = retrieveCommand + " -T extrafiles.txt"
+		retrieveCommand += " -T extrafiles.txt"
 	}
 
 	log := clog.FromContext(ctx)
@@ -498,7 +503,6 @@ func (bw *qemu) GetReleaseData(ctx context.Context, cfg *Config) (*apko_build.Re
 		false,
 		[]string{"sh", "-c", "cat /etc/os-release"},
 	)
-
 	if err != nil {
 		clog.FromContext(ctx).Errorf("failed to get os-release: %v", err)
 		return nil, err
@@ -526,7 +530,6 @@ func getGuestKernelVersion(ctx context.Context, cfg *Config) (version string, er
 		false,
 		[]string{"uname", "-rv"},
 	)
-
 	if err != nil {
 		clog.FromContext(ctx).Errorf("failed to get kernel release version: %v", err)
 		return "", err
@@ -556,7 +559,12 @@ func (b qemuOCILoader) LoadImage(ctx context.Context, layer v1.Layer, arch apko_
 		return "", err
 	}
 	_, err = io.Copy(guestRootfs, layerUncompress)
-	guestRootfs.Close()
+	if closeErr := guestRootfs.Close(); closeErr != nil && err == nil {
+		err = closeErr
+	}
+	if err != nil {
+		return "", err
+	}
 
 	return guestRootfs.Name(), nil
 }
@@ -574,6 +582,12 @@ func createMicroVM(ctx context.Context, cfg *Config) error {
 		return err
 	}
 
+	// Generate VM host key for pre-provisioned, secure host key verification
+	err = generateVMHostKey(ctx, cfg)
+	if err != nil {
+		return err
+	}
+
 	baseargs := []string{}
 	bios := false
 	useVM := false
@@ -673,11 +687,12 @@ func createMicroVM(ctx context.Context, cfg *Config) error {
 	case cfg.CPUModel != "":
 		baseargs = append(baseargs, "-cpu", cfg.CPUModel)
 	case cfg.Arch.ToAPK() != apko_types.ParseArchitecture(runtime.GOARCH).ToAPK():
-		if cfg.Arch.ToAPK() == "aarch64" {
+		switch cfg.Arch.ToAPK() {
+		case "aarch64":
 			baseargs = append(baseargs, "-cpu", "cortex-a76")
-		} else if cfg.Arch.ToAPK() == "x86_64" {
+		case "x86_64":
 			baseargs = append(baseargs, "-cpu", "Haswell-v4")
-		} else {
+		default:
 			return fmt.Errorf("unknown architecture: %s", cfg.Arch.ToAPK())
 		}
 	default:
@@ -707,18 +722,18 @@ func createMicroVM(ctx context.Context, cfg *Config) error {
 	// kill all performances (lots of small files)
 	// instead we will copy back the finished workspace artifacts when done.
 	// this dramatically improves compile time, making them comparable to bwrap or docker runners.
+	// FIXME: set readonly=on once the microvm-init script does not twiddle with /mount/mnt
 	baseargs = append(baseargs, "-fsdev", "local,security_model=mapped,id=fsdev100,path="+cfg.WorkspaceDir)
 	baseargs = append(baseargs, "-device", "virtio-9p-pci,id=fs100,fsdev=fsdev100,mount_tag=defaultshare")
 
 	if cfg.CacheDir != "" {
-		baseargs = append(baseargs, "-fsdev", "local,security_model=mapped,id=fsdev101,path="+cfg.CacheDir)
+		baseargs = append(baseargs, "-fsdev", "local,security_model=mapped,id=fsdev101,readonly=on,path="+cfg.CacheDir)
 		baseargs = append(baseargs, "-device", "virtio-9p-pci,id=fs101,fsdev=fsdev101,mount_tag=melange_cache")
 
 		// ensure the cachedir exists
 		if err := os.MkdirAll(cfg.CacheDir, 0o755); err != nil {
 			return fmt.Errorf("failed to create shared cachedir: %w", err)
 		}
-
 	}
 
 	// if no size is specified, let's go for a default
@@ -739,6 +754,15 @@ func createMicroVM(ctx context.Context, cfg *Config) error {
 		return err
 	}
 
+	// Store initramfs path for cleanup
+	cfg.InitramfsPath = initramFile
+
+	defer func() {
+		if cfg.InitramfsPath != "" {
+			secureDelete(ctx, cfg.InitramfsPath)
+		}
+	}()
+
 	baseargs = append(baseargs, "-kernel", kernelPath)
 	baseargs = append(baseargs, "-initrd", initramFile)
 
@@ -768,6 +792,7 @@ func createMicroVM(ctx context.Context, cfg *Config) error {
 	baseargs = append(baseargs, "-blockdev", "driver=raw,node-name=image.tar,file.driver=file,file.filename="+cfg.ImgRef)
 
 	// qemu-system-x86_64 or qemu-system-aarch64...
+	// #nosec G204 - Architecture is from validated configuration, not user input
 	qemuCmd := exec.CommandContext(ctx, fmt.Sprintf("qemu-system-%s", cfg.Arch.ToAPK()), baseargs...)
 	clog.FromContext(ctx).Info("qemu: starting VM")
 	clog.FromContext(ctx).Debugf("qemu: executing - %s", strings.Join(qemuCmd.Args, " "))
@@ -814,8 +839,8 @@ func createMicroVM(ctx context.Context, cfg *Config) error {
 	qemuExit := make(chan error, 1)
 	go func() {
 		err := qemuCmd.Wait()
-		outWrite.Close()
-		errWrite.Close()
+		_ = outWrite.Close() // Ignore error in goroutine cleanup
+		_ = errWrite.Close() // Ignore error in goroutine cleanup
 		qemuExit <- err
 	}()
 
@@ -848,10 +873,12 @@ func createMicroVM(ctx context.Context, cfg *Config) error {
 	select {
 	case <-started:
 		log.Info("qemu: VM started successfully, SSH server is up")
+		secureDelete(ctx, cfg.InitramfsPath)
+		cfg.InitramfsPath = ""
 	case err := <-qemuExit:
 		defer os.Remove(cfg.ImgRef)
 		defer os.Remove(cfg.Disk)
-		return fmt.Errorf("qemu: VM exited unexpectedly: %v", err)
+		return fmt.Errorf("qemu: VM exited unexpectedly: %w", err)
 	case <-ctx.Done():
 		defer os.Remove(cfg.ImgRef)
 		defer os.Remove(cfg.Disk)
@@ -868,6 +895,10 @@ func createMicroVM(ctx context.Context, cfg *Config) error {
 		return fmt.Errorf("qemu: could not setup SSH client")
 	}
 
+	// Zero out sensitive private key material now that all SSH connections are established
+	// The public key is retained for verification in setupSSHClients
+	zeroSensitiveFields(ctx, cfg)
+
 	stdout, stderr := logwriter.New(log.Info), logwriter.New(log.Warn)
 	defer stdout.Close()
 	defer stderr.Close()
@@ -883,7 +914,7 @@ func createMicroVM(ctx context.Context, cfg *Config) error {
 		setupMountCommand := fmt.Sprintf(
 			"mkdir -p %s %s /mount/upper /mount/work && "+
 				"chmod 1777 /mount/upper && "+
-				"mount -t 9p melange_cache %s && "+
+				"mount -t 9p -o ro melange_cache %s && "+
 				"mount -t overlay overlay -o lowerdir=%s,upperdir=/mount/upper,workdir=/mount/work %s",
 			DefaultCacheDir,
 			filepath.Join("/mount", DefaultCacheDir),
@@ -911,18 +942,22 @@ func createMicroVM(ctx context.Context, cfg *Config) error {
 	}
 
 	clog.FromContext(ctx).Info("qemu: setting up local workspace")
-	// This has to happen as the user we are building as, so
-	// files copied are owned by the build user, and thus cleanup
-	// step that happens at the end of the build to remove the
-	// files succeeds
+	// because microvm-init both mounts and modifies /mnt/ in the
+	// workspace, but want to get rid of that, we need to handle both
+	// remounting and first mount of the defaultshare virtfs share.
+	setupMountCommand := "if grep -q '^defaultshare ' /proc/mounts ; then " +
+		"  mount -t 9p -o remount,ro,trans=virtio defaultshare /mount/mnt ; " +
+		"else " +
+		"  mount -t 9p -o ro,trans=virtio defaultshare /mount/mnt; " +
+		"fi"
 	err = sendSSHCommand(ctx,
-		cfg.SSHBuildClient,
+		cfg.SSHControlClient,
 		cfg,
 		nil,
 		stderr,
 		stdout,
 		false,
-		[]string{"sh", "-c", "find /mnt/ -mindepth 1 -maxdepth 1 -exec cp -a {} /home/build/ \\;"},
+		[]string{"sh", "-c", setupMountCommand},
 	)
 	if err != nil {
 		err = qemuCmd.Process.Kill()
@@ -931,6 +966,55 @@ func createMicroVM(ctx context.Context, cfg *Config) error {
 		}
 	}
 
+	// This has to happen as the user we are building as, so
+	// files copied are owned by the build user, and thus cleanup
+	// step that happens at the end of the build to remove the
+	// files succeeds
+	copyFilesCommand := fmt.Sprintf(
+		"find /mnt/ -mindepth 1 -maxdepth 1 -exec cp -a {} /home/build/ \\; && "+
+			"chown -R %s /home/build",
+		cfg.SSHBuildClient.User(),
+	)
+	err = sendSSHCommand(ctx,
+		cfg.SSHControlBuildClient,
+		cfg,
+		nil,
+		stderr,
+		stdout,
+		false,
+		[]string{"sh", "-c", copyFilesCommand},
+	)
+	if err != nil {
+		err = qemuCmd.Process.Kill()
+		if err != nil {
+			return err
+		}
+	}
+
+	clog.FromContext(ctx).Info("qemu: unmounting host workspace from guest")
+	// now that we've copied files from the host filesystem into the
+	// guest's workspace, we can umount the filesystem, as nothing in
+	// the build should reference the original location and it reduces
+	// the ability of later build steps to possibly interact with the
+	// host.
+	err = sendSSHCommand(ctx,
+		cfg.SSHControlClient,
+		cfg,
+		nil,
+		stderr,
+		stdout,
+		false,
+		[]string{"sh", "-c", "if [ -x /usr/bin/umount ] ; then umount /mount/mnt ; fi"},
+	)
+	if err != nil {
+		clog.FromContext(ctx).Warnf("qemu: failed to unmount /mount/mnt in guest")
+		err = qemuCmd.Process.Kill()
+		if err != nil {
+			return err
+		}
+		// don't fail the build because of this.
+	}
+
 	cfg.QemuPID = qemuCmd.Process.Pid
 	return nil
 }
@@ -1035,7 +1119,7 @@ func getWorkspaceLicenseFiles(ctx context.Context, cfg *Config, extraFiles []str
 		if strings.Contains(f, "melange-out") {
 			continue
 		}
-		if is, _ := license.IsLicenseFile(f); is {
+		if is, _ := license.IsLicenseFile(f, false); is {
 			licenseFiles = append(licenseFiles, f)
 		}
 	}
@@ -1066,11 +1150,11 @@ func streamExtraFilesList(ctx context.Context, cfg *Config, extraFiles []string)
 		session.Stdout = stdout
 		stdin, err := session.StdinPipe()
 		if err != nil {
-			return fmt.Errorf("failed to create stdin pipe: %v", err)
+			return fmt.Errorf("failed to create stdin pipe: %w", err)
 		}
 		cmd := "cat >> /home/build/extrafiles.txt"
 		if err := session.Start(cmd); err != nil {
-			return fmt.Errorf("failed to start command: %v", err)
+			return fmt.Errorf("failed to start command: %w", err)
 		}
 
 		// Write 100 strings there (or remainder)
@@ -1083,15 +1167,15 @@ func streamExtraFilesList(ctx context.Context, cfg *Config, extraFiles []string)
 		if _, err := io.Copy(stdin, strings.NewReader(
 			strings.Join(chunk, "\n")+"\n"),
 		); err != nil {
-			return fmt.Errorf("failed to write content: %v", err)
+			return fmt.Errorf("failed to write content: %w", err)
 		}
 
 		if err := stdin.Close(); err != nil {
-			return fmt.Errorf("failed to close stdin: %v", err)
+			return fmt.Errorf("failed to close stdin: %w", err)
 		}
 
 		if err := session.Wait(); err != nil {
-			return fmt.Errorf("command failed: %v", err)
+			return fmt.Errorf("command failed: %w", err)
 		}
 
 		writtenStrings = endIndex
@@ -1130,7 +1214,7 @@ func generateDiskFile(ctx context.Context, diskSize string) (string, error) {
 	}
 
 	if _, err := os.Stat(diskPath); err != nil {
-		err = os.MkdirAll(diskPath, os.ModePerm)
+		err = os.MkdirAll(diskPath, 0o755)
 		if err != nil {
 			return "", err
 		}
@@ -1148,7 +1232,7 @@ func generateDiskFile(ctx context.Context, diskSize string) (string, error) {
 	}
 
 	// we need bytes
-	size = size * 1024
+	size *= 1024
 
 	clog.FromContext(ctx).Debugf("qemu: generating disk image, name %s, size %s:", diskName.Name(), diskSize)
 	return diskName.Name(), os.Truncate(diskName.Name(), size)
@@ -1187,8 +1271,22 @@ func checkSSHServer(address string) error {
 	return fmt.Errorf("ssh: unknown connection error")
 }
 
+// getHostKey verifies the VM's SSH host key against the pre-provisioned key
 func getHostKey(ctx context.Context, cfg *Config) error {
-	var hostKey ssh.PublicKey
+	if cfg.VMHostKeyPublic == nil {
+		return fmt.Errorf("VM host key not pre-provisioned")
+	}
+
+	clog.FromContext(ctx).Info("qemu: verifying VM host key against pre-provisioned key")
+
+	// Validate this is a localhost connection
+	host, _, err := net.SplitHostPort(cfg.SSHAddress)
+	if err != nil {
+		return fmt.Errorf("invalid SSH address format: %w", err)
+	}
+	if host != "127.0.0.1" && host != "localhost" && host != "::1" {
+		return fmt.Errorf("refusing SSH connection to non-localhost address: %s", host)
+	}
 
 	// default to root user, unless a different user is specified
 	user := "root"
@@ -1205,30 +1303,28 @@ func getHostKey(ctx context.Context, cfg *Config) error {
 		Config: ssh.Config{
 			Ciphers: []string{"aes128-gcm@openssh.com"},
 		},
-		HostKeyCallback: func(hostname string, remote net.Addr, key ssh.PublicKey) error {
-			hostKey = key
-			return nil // Accept the host key for the purpose of retrieving it
-		},
+		HostKeyCallback: ssh.FixedHostKey(cfg.VMHostKeyPublic),
 	}
 
-	// Connect to the SSH server
 	client, err := ssh.Dial("tcp", cfg.SSHAddress, config)
 	if err != nil {
-		clog.FromContext(ctx).Errorf("Failed to dial: %s", err)
-		return err
+		clog.FromContext(ctx).Errorf("Failed to verify and connect to VM: %s", err)
+		return fmt.Errorf("vm host key verification failed (possible security issue): %w", err)
 	}
 	defer client.Close()
 
-	// Write the host key to the known_hosts file
+	clog.FromContext(ctx).Info("qemu: VM host key successfully verified against pre-provisioned key")
+
+	// Write the verified host key to a known_hosts file for use by setupSSHClients
+	// which uses knownhosts.New() for subsequent connections
 	hostKeyLine := fmt.Sprintf("%s %s %s\n%s %s %s\n",
-		cfg.SSHAddress, hostKey.Type(), base64.StdEncoding.EncodeToString(hostKey.Marshal()),
-		cfg.SSHControlAddress, hostKey.Type(), base64.StdEncoding.EncodeToString(hostKey.Marshal()),
+		cfg.SSHAddress, cfg.VMHostKeyPublic.Type(), base64.StdEncoding.EncodeToString(cfg.VMHostKeyPublic.Marshal()),
+		cfg.SSHControlAddress, cfg.VMHostKeyPublic.Type(), base64.StdEncoding.EncodeToString(cfg.VMHostKeyPublic.Marshal()),
 	)
-	clog.FromContext(ctx).Debugf("host-key: %s", hostKeyLine)
 
 	knownHost, err := os.CreateTemp("", "known_hosts_*")
 	if err != nil {
-		clog.FromContext(ctx).Errorf("host-key fetch - failed to create random known_hosts file: %v", err)
+		clog.FromContext(ctx).Errorf("host-key - failed to create known_hosts file: %v", err)
 		return err
 	}
 	defer knownHost.Close()
@@ -1237,16 +1333,18 @@ func getHostKey(ctx context.Context, cfg *Config) error {
 
 	_, err = knownHost.Write([]byte(hostKeyLine))
 	if err != nil {
-		clog.FromContext(ctx).Errorf("host-key fetch - failed to write to known_hosts file: %v", err)
+		clog.FromContext(ctx).Errorf("host-key - failed to write to known_hosts file: %v", err)
 		return err
 	}
+
+	clog.FromContext(ctx).Debugf("qemu: host key written to %s for subsequent connections", cfg.SSHHostKey)
 	return nil
 }
 
 func sendSSHCommand(ctx context.Context, client *ssh.Client,
 	cfg *Config, extraVars map[string]string,
 	stderr, stdout io.Writer,
-	tty bool, command []string,
+	tty bool, command []string, //nolint:unparam
 ) error {
 	// Create a session
 	session, err := client.NewSession()
@@ -1309,11 +1407,11 @@ func sendSSHCommand(ctx context.Context, client *ssh.Client,
 	return nil
 }
 
-func getUserSSHKey() ([]byte, error) {
+func getUserSSHKey(ctx context.Context) ([]byte, error) {
 	socket := os.Getenv("SSH_AUTH_SOCK")
 	conn, err := net.Dial("unix", socket)
 	if err != nil {
-		log.Warnf("Failed to open SSH_AUTH_SOCK: %v, falling back to key search", err)
+		clog.FromContext(ctx).Warnf("Failed to open SSH_AUTH_SOCK: %v, falling back to key search", err)
 		currentUser, err := user.Current()
 		if err != nil {
 			return nil, fmt.Errorf("failed to get current user: %w", err)
@@ -1326,7 +1424,7 @@ func getUserSSHKey() ([]byte, error) {
 
 		for _, keyFile := range knownPublicKeyFiles {
 			path := filepath.Join(sshDir, keyFile)
-			content, err := os.ReadFile(path)
+			content, err := os.ReadFile(path) // #nosec G304 - Reading SSH key from known location
 			if err == nil {
 				return content, nil
 			}
@@ -1376,6 +1474,41 @@ func generateSSHKeys(ctx context.Context, cfg *Config) ([]byte, error) {
 	return ssh.MarshalAuthorizedKey(publicKey), nil
 }
 
+// generateVMHostKey generates an SSH host key pair for the VM.
+func generateVMHostKey(ctx context.Context, cfg *Config) error {
+	clog.FromContext(ctx).Info("qemu: generating SSH host key for VM")
+
+	// Generate Ed25519 key (modern, secure, and fast)
+	publicKey, privateKey, err := ed25519.GenerateKey(rand.Reader)
+	if err != nil {
+		return fmt.Errorf("failed to generate VM host key: %w", err)
+	}
+
+	// Create SSH signer from the private key
+	signer, err := ssh.NewSignerFromKey(privateKey)
+	if err != nil {
+		return fmt.Errorf("failed to create signer from VM host key: %w", err)
+	}
+
+	// Marshal the private key in OpenSSH format for injection into initramfs
+	privateKeyPEM, err := ssh.MarshalPrivateKey(privateKey, "")
+	if err != nil {
+		return fmt.Errorf("failed to marshal VM host private key: %w", err)
+	}
+
+	// Store the signer, public key, raw private key, and marshaled private key
+	cfg.VMHostKeySigner = signer
+	cfg.VMHostKeyPublic = signer.PublicKey()
+	cfg.VMHostKeyPrivate = privateKey // Keep reference to raw key for explicit zeroing
+	cfg.VMHostKeyPrivateKeyBytes = pem.EncodeToMemory(privateKeyPEM)
+
+	clog.FromContext(ctx).Debugf("qemu: generated VM host key: %s %s",
+		cfg.VMHostKeyPublic.Type(),
+		base64.StdEncoding.EncodeToString(publicKey))
+
+	return nil
+}
+
 // Use binary values for all units
 const (
 	_  = iota
@@ -1487,13 +1620,16 @@ func getAvailableMemoryKB() int {
 		}
 
 		sysctlOutStr := strings.TrimSpace(string(sysctlOut))
-		memSize, err = strconv.ParseInt(sysctlOutStr, 10, 64)
+		memSize, err = strconv.ParseInt(sysctlOutStr, 10, strconv.IntSize)
 		if err != nil {
 			return mem
 		}
 
 		// use at most 50% of total ram, in kb
-		return int(memSize) / 2 / 1024
+		if memSize > 0 {
+			return int(memSize) / 2 / 1024
+		}
+		return mem
 	}
 
 	return mem
@@ -1509,6 +1645,90 @@ func randomPortN() (int, error) {
 	return l.Addr().(*net.TCPAddr).Port, nil
 }
 
+// zeroSensitiveFields zeros out sensitive cryptographic material from memory after it's no longer needed.
+func zeroSensitiveFields(ctx context.Context, cfg *Config) {
+	clog.FromContext(ctx).Debug("qemu: zeroing sensitive cryptographic material from memory")
+
+	if cfg.VMHostKeyPrivate != nil {
+		for i := range cfg.VMHostKeyPrivate {
+			cfg.VMHostKeyPrivate[i] = 0
+		}
+		cfg.VMHostKeyPrivate = nil
+	}
+
+	if cfg.VMHostKeyPrivateKeyBytes != nil {
+		for i := range cfg.VMHostKeyPrivateKeyBytes {
+			cfg.VMHostKeyPrivateKeyBytes[i] = 0
+		}
+		cfg.VMHostKeyPrivateKeyBytes = nil
+	}
+
+	cfg.VMHostKeySigner = nil
+
+	runtime.GC()
+
+	clog.FromContext(ctx).Debug("qemu: sensitive key material zeroed and GC triggered")
+}
+
+// secureDelete overwrites a file with random data before deleting it.
+func secureDelete(ctx context.Context, path string) {
+	if path == "" {
+		return
+	}
+
+	info, err := os.Stat(path)
+	if err != nil {
+		if !os.IsNotExist(err) {
+			clog.FromContext(ctx).Debugf("secure delete: cannot stat %s: %v", path, err)
+		}
+		return
+	}
+
+	if err := os.Chmod(path, 0o600); err != nil {
+		clog.FromContext(ctx).Debugf("secure delete: cannot make %s writable: %v", path, err)
+	}
+
+	file, err := os.OpenFile(path, os.O_WRONLY, 0) // #nosec G304 - Secure file deletion operation
+	if err != nil {
+		clog.FromContext(ctx).Warnf("secure delete: cannot open %s for overwriting: %v", path, err)
+		_ = os.Remove(path) // Best effort cleanup in error path
+		return
+	}
+	defer file.Close()
+
+	size := info.Size()
+	buf := make([]byte, 4096)
+	for written := int64(0); written < size; {
+		toWrite := min(int64(len(buf)), size-written)
+
+		if _, err := rand.Read(buf[:toWrite]); err != nil {
+			clog.FromContext(ctx).Warnf("secure delete: failed to generate random data: %v", err)
+			break
+		}
+
+		if _, err := file.Write(buf[:toWrite]); err != nil {
+			clog.FromContext(ctx).Warnf("secure delete: failed to overwrite %s: %v", path, err)
+			break
+		}
+
+		written += toWrite
+	}
+
+	if err := file.Sync(); err != nil {
+		clog.FromContext(ctx).Warnf("secure delete: failed to sync %s: %v", path, err)
+	}
+	// Close explicitly before remove (defer also closes but this ensures it's closed first)
+	if err := file.Close(); err != nil {
+		clog.FromContext(ctx).Warnf("secure delete: failed to close %s: %v", path, err)
+	}
+
+	if err := os.Remove(path); err != nil {
+		clog.FromContext(ctx).Warnf("secure delete: failed to remove %s: %v", path, err)
+	} else {
+		clog.FromContext(ctx).Debugf("secure delete: successfully deleted %s", path)
+	}
+}
+
 func generateCpio(ctx context.Context, cfg *Config) (string, error) {
 	/*
 	 * we only build once, useful for local development, we
@@ -1521,21 +1741,32 @@ func generateCpio(ctx context.Context, cfg *Config) (string, error) {
 
 	cacheDir = filepath.Join(cacheDir, "melange-cpio")
 
-	initramfs := filepath.Join(
+	baseInitramfs := filepath.Join(
 		cacheDir,
 		"melange-guest.initramfs.cpio")
-	initramfsInfo, err := os.Stat(initramfs)
+	initramfsInfo, err := os.Stat(baseInitramfs)
 
-	// if file is presend and less than 24h old, then we just reuse it
-	if err == nil && time.Since(initramfsInfo.ModTime()) < 24*time.Hour {
-		return initramfs, nil
+	// Check if we can use the cached base initramfs (less than 24h old)
+	useCache := err == nil && time.Since(initramfsInfo.ModTime()) < 24*time.Hour
+
+	if !useCache {
+		err = generateBaseInitramfs(ctx, cfg, baseInitramfs, cacheDir)
+		if err != nil {
+			return "", err
+		}
 	}
 
-	clog.FromContext(ctx).Info("qemu: generating initramfs")
+	return injectHostKeyIntoCpio(ctx, cfg, baseInitramfs)
+}
 
-	err = os.MkdirAll(cacheDir, os.ModePerm)
+// generateBaseInitramfs creates the base initramfs with microvm-init.
+// This is cached for performance as it doesn't change often.
+func generateBaseInitramfs(ctx context.Context, cfg *Config, initramfsPath, cacheDir string) error {
+	clog.FromContext(ctx).Info("qemu: generating base initramfs")
+
+	err := os.MkdirAll(cacheDir, 0o755)
 	if err != nil {
-		return "", fmt.Errorf("unable to dest directory: %w", err)
+		return fmt.Errorf("unable to create dest directory: %w", err)
 	}
 
 	spec := apko_types.ImageConfiguration{
@@ -1555,26 +1786,26 @@ func generateCpio(ctx context.Context, cfg *Config) (string, error) {
 
 	tmpDir, err := os.MkdirTemp("", "melange-guest-*.initramfs")
 	if err != nil {
-		return "", fmt.Errorf("unable to create build context: %w", err)
+		return fmt.Errorf("unable to create build context: %w", err)
 	}
 	defer os.RemoveAll(tmpDir)
 
 	bc, err := apko_build.New(ctx, apkofs.DirFS(ctx, tmpDir, apkofs.WithCreateDir()), opts...)
 	if err != nil {
-		return "", fmt.Errorf("unable to create build context: %w", err)
+		return fmt.Errorf("unable to create build context: %w", err)
 	}
 
 	bc.Summarize(ctx)
 	if err := bc.BuildImage(ctx); err != nil {
-		return "", fmt.Errorf("unable to generate image: %w", err)
+		return fmt.Errorf("unable to generate image: %w", err)
 	}
 	layerTarGZ, layer, err := bc.ImageLayoutToLayer(ctx)
 	if err != nil {
-		return "", err
+		return err
 	}
 	defer os.Remove(layerTarGZ)
 
-	log.Debugf("using %s for image layer", layerTarGZ)
+	clog.FromContext(ctx).Debugf("using %s for image layer", layerTarGZ)
 
 	// in case of some kernel images, we also need the /lib/modules directory to load
 	// necessary drivers, like 9p, virtio_net which are foundamental for the VM working.
@@ -1585,23 +1816,95 @@ func generateCpio(ctx context.Context, cfg *Config) (string, error) {
 			layer, err = injectKernelModules(ctx, layer, qemuModule)
 			if err != nil {
 				clog.FromContext(ctx).Errorf("qemu: could not inject needed kernel modules into initramfs: %v", err)
-				return "", err
+				return err
 			}
 		}
 	}
 
-	guestInitramfs, err := os.Create(initramfs)
+	guestInitramfs, err := os.Create(initramfsPath) // #nosec G304 - Creating initramfs in build output
 	if err != nil {
 		clog.FromContext(ctx).Errorf("failed to create cpio initramfs: %v", err)
-		return "", err
+		return err
 	}
+	defer guestInitramfs.Close()
 
 	if err := apko_cpio.FromLayer(layer, guestInitramfs); err != nil {
 		clog.FromContext(ctx).Errorf("failed to convert cpio initramfs: %v", err)
-		return "", err
+		return err
 	}
 
-	return guestInitramfs.Name(), nil
+	return nil
+}
+
+// injectHostKeyIntoCpio creates a new initramfs by concatenating the base
+// initramfs with a secondary cpio containing the VM's SSH host key.
+// This allows us to use cached base initramfs while injecting fresh keys each time.
+//
+// The file exists only briefly until QEMU reads it into memory, at which point it should be deleted.
+func injectHostKeyIntoCpio(ctx context.Context, cfg *Config, baseInitramfs string) (string, error) {
+	if cfg.VMHostKeySigner == nil || cfg.VMHostKeyPrivateKeyBytes == nil {
+		return "", fmt.Errorf("VM host key not generated")
+	}
+
+	clog.FromContext(ctx).Debug("qemu: injecting SSH host key into initramfs")
+
+	combinedInitramfs, err := os.CreateTemp("", "melange-initramfs-*.cpio")
+	if err != nil {
+		return "", fmt.Errorf("failed to create temp initramfs file: %w", err)
+	}
+	defer combinedInitramfs.Close()
+	initramfsPath := combinedInitramfs.Name()
+
+	success := false
+	defer func() {
+		if !success {
+			secureDelete(ctx, initramfsPath)
+		}
+	}()
+
+	baseFile, err := os.Open(baseInitramfs) // #nosec G304 - Reading base initramfs from cache
+	if err != nil {
+		return "", fmt.Errorf("failed to open base initramfs: %w", err)
+	}
+	defer baseFile.Close()
+
+	if _, err := io.Copy(combinedInitramfs, baseFile); err != nil {
+		return "", fmt.Errorf("failed to copy base initramfs: %w", err)
+	}
+
+	// Create u-root CPIO writer for appending additional files
+	cpioWriter := cpio.Newc.Writer(combinedInitramfs)
+
+	publicKeyBytes := ssh.MarshalAuthorizedKey(cfg.VMHostKeyPublic)
+
+	// Write the private key file with 0400 permissions (read-only, owner only)
+	privateKeyRecord := cpio.StaticFile("etc/ssh/ssh_host_ed25519_key", string(cfg.VMHostKeyPrivateKeyBytes), 0o400)
+	if err := cpio.WriteRecordsAndDirs(cpioWriter, []cpio.Record{privateKeyRecord}); err != nil {
+		return "", fmt.Errorf("failed to write host private key: %w", err)
+	}
+
+	// Write the public key file with 0644 permissions (world-readable)
+	publicKeyRecord := cpio.StaticFile("etc/ssh/ssh_host_ed25519_key.pub", string(publicKeyBytes), 0o644)
+	if err := cpio.WriteRecordsAndDirs(cpioWriter, []cpio.Record{publicKeyRecord}); err != nil {
+		return "", fmt.Errorf("failed to write host public key: %w", err)
+	}
+
+	// Write CPIO trailer to finalize the archive
+	if err := cpioWriter.WriteRecord(cpio.TrailerRecord); err != nil {
+		return "", fmt.Errorf("failed to finalize cpio: %w", err)
+	}
+
+	if err := combinedInitramfs.Close(); err != nil {
+		return "", fmt.Errorf("failed to close initramfs: %w", err)
+	}
+
+	if err := os.Chmod(initramfsPath, 0o400); err != nil {
+		return "", fmt.Errorf("failed to set read-only permissions on initramfs: %w", err)
+	}
+
+	clog.FromContext(ctx).Debugf("qemu: created initramfs with injected host key at %s", initramfsPath)
+	success = true
+	return initramfsPath, nil
 }
 
 // in case of external modules (usually for 9p and virtio) we need a matching /lib/modules/kernel-$(uname)
@@ -1632,6 +1935,7 @@ func injectKernelModules(ctx context.Context, rootfs v1.Layer, modulesPath strin
 		if err := tarWriter.WriteHeader(header); err != nil {
 			return nil, err
 		}
+		// #nosec G110 - Copying trusted tar content in controlled build environment
 		if _, err := io.Copy(tarWriter, tartReader); err != nil {
 			return nil, err
 		}
@@ -1646,7 +1950,7 @@ func injectKernelModules(ctx context.Context, rootfs v1.Layer, modulesPath strin
 			return nil
 		}
 
-		data, err := os.ReadFile(path)
+		data, err := os.ReadFile(path) // #nosec G304 - Reading guest file for verification
 		if err != nil {
 			return fmt.Errorf("failed to read file %s: %w", path, err)
 		}
