@@ -38,7 +38,13 @@ import (
 	"chainguard.dev/melange/pkg/config"
 )
 
-var libDirs = []string{"lib/", "usr/lib/", "lib64/", "usr/lib64/"}
+// LibDirs is the list of library directories to search for shared objects.
+// This is exported so that callers can append to it as needed.
+var LibDirs = []string{"lib/", "usr/lib/", "lib64/", "usr/lib64/"}
+
+// BinDirs is the list of binary directories to search for commands.
+// This is exported so that callers can append to it as needed.
+var BinDirs = []string{"bin/", "sbin/", "usr/bin/", "usr/sbin/"}
 
 // SCAFS represents the minimum required filesystem accessors which are needed by
 // the SCA engine.
@@ -85,14 +91,21 @@ type SCAHandle interface {
 }
 
 // DependencyGenerator takes an SCAHandle, config.Dependencies pointer
-// and a list of paths to be appended to libDirs and returns findings
+// and a list of paths to be appended to LibDirs and returns findings
 // based on analysis.
 type DependencyGenerator func(context.Context, SCAHandle, *config.Dependencies, []string) error
 
 func isInDir(path string, dirs []string) bool {
 	mydir := filepath.Dir(path)
 	for _, d := range dirs {
-		if mydir == d || mydir+"/" == d {
+		// allow the /... suffix (i.e. "usr/...") to indicate recursive matching
+		if prefix, ok := strings.CutSuffix(d, "/..."); ok {
+			// Ensure we match the directory exactly or as a prefix followed by /
+			// This prevents /dir matching /dirother
+			if mydir == prefix || strings.HasPrefix(mydir, prefix+"/") {
+				return true
+			}
+		} else if mydir == d || mydir+"/" == d {
 			return true
 		}
 	}
@@ -133,12 +146,7 @@ func isHostProvidedLibrary(lib string) bool {
 		"libnvoptix.so.1",
 	}
 
-	for _, hostLib := range hostLibs {
-		if lib == hostLib {
-			return true
-		}
-	}
-	return false
+	return slices.Contains(hostLibs, lib)
 }
 
 // getLdSoConfDLibPaths will iterate over the files being installed by
@@ -198,7 +206,7 @@ func getLdSoConfDLibPaths(ctx context.Context, hdl SCAHandle) ([]string, error)
 					continue
 				}
 				// Strip the initial slash since
-				// libDirs paths need to be relative.
+				// LibDirs paths need to be relative.
 				line = line[1:]
 				log.Infof("    found extra lib path %s", line)
 				extraLibPaths = append(extraLibPaths, line)
@@ -242,7 +250,7 @@ func generateCmdProviders(ctx context.Context, hdl SCAHandle, generated *config.
 		}
 
 		if mode.Perm()&0o555 == 0o555 {
-			if isInDir(path, []string{"bin/", "sbin/", "usr/bin/", "usr/sbin/"}) {
+			if isInDir(path, BinDirs) {
 				basename := filepath.Base(path)
 				log.Infof("  found command %s", path)
 				generated.Provides = append(generated.Provides, fmt.Sprintf("cmd:%s=%s", basename, hdl.Version()))
@@ -295,7 +303,8 @@ func dereferenceCrossPackageSymlink(hdl SCAHandle, path string, extraLibDirs []s
 
 	realPath = filepath.Base(realPath)
 
-	expandedLibDirs := append(libDirs, extraLibDirs...)
+	expandedLibDirs := append([]string{}, LibDirs...)
+	expandedLibDirs = append(expandedLibDirs, extraLibDirs...)
 
 	for _, pkgName := range targetPackageNames {
 		baseFS, err := hdl.FilesystemForRelative(pkgName)
@@ -558,7 +567,8 @@ func generateSharedObjectNameDeps(ctx context.Context, hdl SCAHandle, generated
 		return err
 	}
 
-	expandedLibDirs := append(libDirs, extraLibDirs...)
+	expandedLibDirs := append([]string{}, LibDirs...)
+	expandedLibDirs = append(expandedLibDirs, extraLibDirs...)
 
 	if err := fs.WalkDir(fsys, ".", func(path string, d fs.DirEntry, err error) error {
 		if err != nil {
@@ -946,7 +956,6 @@ func generateDocDeps(ctx context.Context, hdl SCAHandle, generated *config.Depen
 		}
 
 		if isInDir(path, []string{"usr/share/man"}) {
-
 			// Do not add a man-db dependency if one already exists.
 			for _, dep := range hdl.BaseDependencies().Runtime {
 				if dep == "man-db" {
@@ -959,7 +968,6 @@ func generateDocDeps(ctx context.Context, hdl SCAHandle, generated *config.Depen
 		}
 
 		if isInDir(path, []string{"usr/share/info"}) {
-
 			// Do not add a texinfo dependency if one already exists.
 			for _, dep := range hdl.BaseDependencies().Runtime {
 				if dep == "texinfo" {
@@ -1007,13 +1015,14 @@ func getShbang(fp io.Reader) (string, error) {
 
 	buf := make([]byte, 80)
 	blen, err := io.ReadFull(fp, buf)
-	if err == io.EOF {
+	switch {
+	case errors.Is(err, io.EOF):
 		return "", nil
-	} else if err == io.ErrUnexpectedEOF {
+	case errors.Is(err, io.ErrUnexpectedEOF):
 		if blen < 2 {
 			return "", nil
 		}
-	} else if err != nil {
+	case err != nil:
 		return "", err
 	}
 
@@ -1031,15 +1040,14 @@ func getShbang(fp io.Reader) (string, error) {
 
 	// if #! is '/usr/bin/env foo', then use next arg as the dep
 	if bin == "/usr/bin/env" {
-		if len(toks) == 1 {
+		switch {
+		case len(toks) == 1:
 			return "", fmt.Errorf("a shbang of only '/usr/bin/env'")
-		} else if len(toks) == 2 {
+		case len(toks) == 2:
 			bin = toks[1]
-		} else if len(toks) >= 3 && toks[1] == "-S" && !strings.HasPrefix(toks[2], "-") {
-			// we really need a env argument parser to figure out what the next cmd is.
-			// special case handle /usr/bin/env -S prog [arg1 [arg2 [...]]]
+		case len(toks) >= 3 && toks[1] == "-S" && !strings.HasPrefix(toks[2], "-"):
 			bin = toks[2]
-		} else {
+		default:
 			return "", fmt.Errorf("a shbang of only '/usr/bin/env' with multiple arguments (%d %s)", len(toks), strings.Join(toks, " "))
 		}
 	}
@@ -1087,7 +1095,9 @@ func generateShbangDeps(ctx context.Context, hdl SCAHandle, generated *config.De
 			} else if shbang != "" {
 				cmds[filepath.Base(shbang)] = path
 			}
-			fp.Close()
+			if err := fp.Close(); err != nil {
+				log.Warnf("Error closing %s: %v", path, err)
+			}
 		} else {
 			log.Infof("Failed to open %s: %v", path, err)
 		}
